; -*- Lisp -*-
; lambda-calculus-interpreter: an EuLisp library for interpretation of
; commands used for manipulating lambda-calculus
;
; Copyright (c) 2001 Ian Hickson <ian@hixie.ch>
;
; This program is free software; you can redistribute it and/or modify
; it under the terms of the GNU General Public License as published by
; the Free Software Foundation; either version 2 of the License, or
; (at your option) any later version.
;
; This program is distributed in the hope that it will be useful, but
; WITHOUT ANY WARRANTY; without even the implied warranty of
; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
; General Public License for more details.
;
; You should have received a copy of the GNU General Public License
; along with this program; if not, write to the Free Software
; Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

(defmodule lambda-calculus-interpreter
  (import (level0 lambda-calculus lambda-calculus-tokeniser lambda-calculus-parser utilities))

  ;; error that can be raised by the interface
  (defcondition <interpreter-condition> <condition>)
  (defcondition <interpreter-unknown-command-error> <interpreter-condition>)
  (defcondition <interpreter-wrong-arguments-error> <interpreter-condition>)
  (defcondition <interpreter-reduction-aborted> <interpreter-condition>)
  (defcondition <thread-abort-condition> <thread-condition>)
  (defcondition <thread-other-thread-error-condition> <thread-condition> propagated-error ())

  ;; the main command executor
  (defun execute (command references echo prompt)
    (cond ((eql (size command) 0) ()) ; blank line
          ((equal command "help") (help))
          ((equal command "warranty") (warranty))
          ((equal command "license") (license))
          ((equal (element command 0) #\#) ()) ; skip comments
          (t (let ((reference ())
                   (start 0))
               ; see if we have to set a reference to the value of the command
               (if (and (> (size command) 2) (equal (element command 1) #\=))
                   (progn (setq reference (element command 0))
                          (setq start 2)))
               ; parse the command
               (let* ((words (split command #\space equal start))
                      ; execute it
                      (result (dispatch words references echo prompt)))
                 ; store the result in the reference if appropriate
                 (if reference 
                     ((setter element) references reference result))
                 ; print the result if appropriate
                 (if echo
                     (print result))
                 ; and finally return it
                 result)))))

  (defun dispatch (words references echo prompt)
    (let* ((word-count (size words))
           (command (and (> word-count 0) (element words 0)))
           (arg1 (and (> word-count 1) (element words 1)))
           (arg2 (and (> word-count 2) (element words 2))))
      (cond ((equal command "step") (do-step arg1 references))
            ((equal command "reduce") (do-reduce arg1 references prompt))
            ((equal command "copy") (do-copy arg1 references))
            ((equal command "print") (do-print arg1 references))
            ((equal command "variable") (do-variable (convert arg1 <number>) arg2 references))
            ((equal command "rename") (do-rename arg1 arg2 references))
            ((equal command "variables") (do-variables arg1 references))
            ((and (equal word-count 1) (> (size command) 5)
                  (eql (element command 0) #\[)
                  (eql (element command 2) #\/)
                  (eql (element command 4) #\]))
             (do-substitution (element command 1)
                              (element command 3)
                              (substring command 5 (size command))
                              references echo))
            ((equal word-count 1) (do-nothing command references))
            (t (error "unknown command -- type 'help' for a list of allowed commands" <interpreter-unknown-command-error>)))))

  (defun evaluate (expression references)
    (if expression 
        (or (if (characterp expression)
                (element references expression)
              (and (eql (size expression) 1)
                   (element references (element expression 0))))
            (parse (tokenise expression)))
      (error "missing argument" <interpreter-wrong-arguments-error>)))

  (defun do-step (expression references)
    (beta-reduce-step (evaluate expression references)))

  (defun reducing-thread (term parent-thread)
    (let/cc failed
     (with-handler ; report any error to parent thread
      (generic-lambda ((condition <condition>) (resume <function>))
                      method: (((condition <condition>) (resume <function>))
                               ; don't ask me why the 'message' and 'propagated-error' keywords in the next line have colons
                               ; XXX per EuLisp spec we don't need them, but in euscheme they seem to be required.
                               (signal (make <thread-other-thread-error-condition> message: "an error occured during beta reduction" propagated-error: condition) () parent-thread)
                               (failed)))
      (beta-reduce term))))

  (defun try-to-reduce (thread prompt)
    (if (wait thread (* ticks-per-second 5))
        ; w00t, got a value
        (thread-value thread)
      ; still running, prompt user about what to do
      ; note that reduction carries on running in background while the prompt is up
      (if (prompt "After five seconds, the reduction had not completed. Abort [y/N]? " #\Y #\y)
          ; user said to abort
          (progn (signal (make <thread-abort-condition>) () thread) ; kill other thread too
                 (error "beta reduction did not terminate" <interpreter-reduction-aborted>))
        ; loop
        (try-to-reduce thread prompt))))

  (defun do-reduce (expression references prompt)
    (let ((term (evaluate expression references))
          ; XXX Per EuLisp we should be able to instantiate a
          ; <thread>, but since in euscheme <thread> is abstract, we
          ; instantiate a <simple-thread> instead.
          ; XXX keyword 'function:' should really be 'init-function'
          ; according to EuLisp.
          (thread (make <simple-thread> function: reducing-thread)))
      (try-to-reduce (thread-start thread term (current-thread)) prompt)))

  (defun do-print (expression references)
    (let ((term (evaluate expression references)))
      (print term)
      term))

  (defun do-copy (expression references)
    (deep-copy (evaluate expression references)))

  (defun do-variable (index expression references)
    (let* ((term (evaluate expression references))
           (variables (get-free-variables term)))
      (if (< index (size variables))
          (element variables index)
        (error "argument out of range" <interpreter-wrong-arguments-error>))))

  (defun do-rename (variable new-name references)
    (let ((term (evaluate variable references)))
      (if (variablep variable) ; term must be <variable>
          (progn (alpha-rename variable new-name)
                 variable)
        (error "argument must be a variable" <interpreter-wrong-arguments-error>))))

  (defun do-variables (expression references)
    (let* ((term (evaluate expression references))
           (variables (get-free-variables term)))
      (print-variables variables 0)
      term))

  (defun print-variables (variables index)
    (if (not (eq variables ()))
        (progn (print (concatenate "variable " (convert index <string>) ": " (convert (car variables) <string>)))
               (print-variables (cdr variables) (+ index 1)))))

  (defun do-substitution (replace search expression references echo)
    (let ((term (evaluate expression references))
          (search-term (evaluate search references))
          (replace-term (evaluate replace references)))
      (if (variablep search-term) ; search-term must be <variable>
          (let* ((new-term (substitute term search-term replace-term))
                 (clashes (clean-name-clashes new-term)))
            (if echo 
                (print-clashes clashes))
            new-term)
        (error "argument must be a variable" <interpreter-wrong-arguments-error>))))

  (defun print-clashes (clashes)
    (if (not (eq clashes ()))
        (progn (print (concatenate "variable " (convert (car (car clashes)) <string>) " was renamed to " (convert (cdr (car clashes)) <string>) "."))
               (print-clashes (cdr clashes)))))

  (defun do-nothing (expression references)
    (evaluate expression references))


  ;;; and now the boring functions that just call (print).

  (defun help ()
    (print "Lambda Calculus
The following commands are supported:
  S             returns S.
  [R/V]S        returns a new expression consisting of S, where every
                mention of variable V is replaced by R. The variables
                are then renamed to remove any name clashes in the new
                expression. (This may affect variables in other
                expressions, such as S itself.)
  reduce S      returns a new expression that is the beta reduced form
                of S. Since this may not terminate, you may be
                prompted to abort the reduction in interactive mode.
                In batch mode care should be taken to not reduce
                expressions that do not terminate.
  step S        returns a new expression that is S, beta reduced once.
  print S       prints and returns S.
  copy S        returns a deep copy of S.
  variables S   lists the unbound variables in S and returns S.
  variable n S  returns the nth unbound variable in S.
  rename V a    alpha-renames V to a.
where:
  S is any Lambda Calculus expression or reference, e.g. R or (Lx.yz)a
  R is any Lambda Calculus reference, e.g. R
  V is any Lambda Calculus reference that is a variable, e.g. R
  n is a number, and a is a letter.

Any command may be prefixed by R= to assign the result of the command
to R. Note that the variables used in lambda expressions are in a
different namespace to the references. Also bear in mind that two
variables with the same name may not actually be the same variable if
they were not created in the same expression.

Lines prefixed with a '#' symbol are ignored.

Examples:
  > A=Lx.xy
  Lx.xy
  > B=aa
  aa
  > variables B
  variable 0: a
  aa
  > a=variable 0 B
  a
  > B=[A/a]B
  (Lx.xy)(Lx.xy)
  > reduce B
  yy
  > B
  (Lx.yy)(Lx.yy)

To exit the Lambda Calculus interpreter, type: quit
"))

  (defun license ()
    (print "		    GNU GENERAL PUBLIC LICENSE
   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION

  0. This License applies to any program or other work which contains
a notice placed by the copyright holder saying it may be distributed
under the terms of this General Public License.  The \"Program\", below,
refers to any such program or work, and a \"work based on the Program\"
means either the Program or any derivative work under copyright law:
that is to say, a work containing the Program or a portion of it,
either verbatim or with modifications and/or translated into another
language.  (Hereinafter, translation is included without limitation in
the term \"modification\".)  Each licensee is addressed as \"you\".

Activities other than copying, distribution and modification are not
covered by this License; they are outside its scope.  The act of
running the Program is not restricted, and the output from the Program
is covered only if its contents constitute a work based on the
Program (independent of having been made by running the Program).
Whether that is true depends on what the Program does.")
    (print "
  1. You may copy and distribute verbatim copies of the Program's
source code as you receive it, in any medium, provided that you
conspicuously and appropriately publish on each copy an appropriate
copyright notice and disclaimer of warranty; keep intact all the
notices that refer to this License and to the absence of any warranty;
and give any other recipients of the Program a copy of this License
along with the Program.

You may charge a fee for the physical act of transferring a copy, and
you may at your option offer warranty protection in exchange for a fee.")
    (print "
  2. You may modify your copy or copies of the Program or any portion
of it, thus forming a work based on the Program, and copy and
distribute such modifications or work under the terms of Section 1
above, provided that you also meet all of these conditions:

    a) You must cause the modified files to carry prominent notices
    stating that you changed the files and the date of any change.

    b) You must cause any work that you distribute or publish, that in
    whole or in part contains or is derived from the Program or any
    part thereof, to be licensed as a whole at no charge to all third
    parties under the terms of this License.

    c) If the modified program normally reads commands interactively
    when run, you must cause it, when started running for such
    interactive use in the most ordinary way, to print or display an
    announcement including an appropriate copyright notice and a
    notice that there is no warranty (or else, saying that you provide
    a warranty) and that users may redistribute the program under
    these conditions, and telling the user how to view a copy of this
    License.  (Exception: if the Program itself is interactive but
    does not normally print such an announcement, your work based on
    the Program is not required to print an announcement.)

These requirements apply to the modified work as a whole.  If
identifiable sections of that work are not derived from the Program,
and can be reasonably considered independent and separate works in
themselves, then this License, and its terms, do not apply to those
sections when you distribute them as separate works.  But when you
distribute the same sections as part of a whole which is a work based
on the Program, the distribution of the whole must be on the terms of
this License, whose permissions for other licensees extend to the
entire whole, and thus to each and every part regardless of who wrote it.

Thus, it is not the intent of this section to claim rights or contest
your rights to work written entirely by you; rather, the intent is to
exercise the right to control the distribution of derivative or
collective works based on the Program.

In addition, mere aggregation of another work not based on the Program
with the Program (or with a work based on the Program) on a volume of
a storage or distribution medium does not bring the other work under
the scope of this License.")
    (print "
  3. You may copy and distribute the Program (or a work based on it,
under Section 2) in object code or executable form under the terms of
Sections 1 and 2 above provided that you also do one of the following:

    a) Accompany it with the complete corresponding machine-readable
    source code, which must be distributed under the terms of Sections
    1 and 2 above on a medium customarily used for software interchange; or,

    b) Accompany it with a written offer, valid for at least three
    years, to give any third party, for a charge no more than your
    cost of physically performing source distribution, a complete
    machine-readable copy of the corresponding source code, to be
    distributed under the terms of Sections 1 and 2 above on a medium
    customarily used for software interchange; or,

    c) Accompany it with the information you received as to the offer
    to distribute corresponding source code.  (This alternative is
    allowed only for noncommercial distribution and only if you
    received the program in object code or executable form with such
    an offer, in accord with Subsection b above.)

The source code for a work means the preferred form of the work for
making modifications to it.  For an executable work, complete source
code means all the source code for all modules it contains, plus any
associated interface definition files, plus the scripts used to
control compilation and installation of the executable.  However, as a
special exception, the source code distributed need not include
anything that is normally distributed (in either source or binary
form) with the major components (compiler, kernel, and so on) of the
operating system on which the executable runs, unless that component
itself accompanies the executable.

If distribution of executable or object code is made by offering
access to copy from a designated place, then offering equivalent
access to copy the source code from the same place counts as
distribution of the source code, even though third parties are not
compelled to copy the source along with the object code.")
    (print "
  4. You may not copy, modify, sublicense, or distribute the Program
except as expressly provided under this License.  Any attempt
otherwise to copy, modify, sublicense or distribute the Program is
void, and will automatically terminate your rights under this License.
However, parties who have received copies, or rights, from you under
this License will not have their licenses terminated so long as such
parties remain in full compliance.")
    (print "
  5. You are not required to accept this License, since you have not
signed it.  However, nothing else grants you permission to modify or
distribute the Program or its derivative works.  These actions are
prohibited by law if you do not accept this License.  Therefore, by
modifying or distributing the Program (or any work based on the
Program), you indicate your acceptance of this License to do so, and
all its terms and conditions for copying, distributing or modifying
the Program or works based on it.")
    (print "
  6. Each time you redistribute the Program (or any work based on the
Program), the recipient automatically receives a license from the
original licensor to copy, distribute or modify the Program subject to
these terms and conditions.  You may not impose any further
restrictions on the recipients' exercise of the rights granted herein.
You are not responsible for enforcing compliance by third parties to
this License.")
    (print "
  7. If, as a consequence of a court judgment or allegation of patent
infringement or for any other reason (not limited to patent issues),
conditions are imposed on you (whether by court order, agreement or
otherwise) that contradict the conditions of this License, they do not
excuse you from the conditions of this License.  If you cannot
distribute so as to satisfy simultaneously your obligations under this
License and any other pertinent obligations, then as a consequence you
may not distribute the Program at all.  For example, if a patent
license would not permit royalty-free redistribution of the Program by
all those who receive copies directly or indirectly through you, then
the only way you could satisfy both it and this License would be to
refrain entirely from distribution of the Program.

If any portion of this section is held invalid or unenforceable under
any particular circumstance, the balance of the section is intended to
apply and the section as a whole is intended to apply in other
circumstances.

It is not the purpose of this section to induce you to infringe any
patents or other property right claims or to contest validity of any
such claims; this section has the sole purpose of protecting the
integrity of the free software distribution system, which is
implemented by public license practices.  Many people have made
generous contributions to the wide range of software distributed
through that system in reliance on consistent application of that
system; it is up to the author/donor to decide if he or she is willing
to distribute software through any other system and a licensee cannot
impose that choice.

This section is intended to make thoroughly clear what is believed to
be a consequence of the rest of this License.")
    (print "
  8. If the distribution and/or use of the Program is restricted in
certain countries either by patents or by copyrighted interfaces, the
original copyright holder who places the Program under this License
may add an explicit geographical distribution limitation excluding
those countries, so that distribution is permitted only in or among
countries not thus excluded.  In such case, this License incorporates
the limitation as if written in the body of this License.")
    (print "
  9. The Free Software Foundation may publish revised and/or new versions
of the General Public License from time to time.  Such new versions will
be similar in spirit to the present version, but may differ in detail to
address new problems or concerns.

Each version is given a distinguishing version number.  If the Program
specifies a version number of this License which applies to it and \"any
later version\", you have the option of following the terms and conditions
either of that version or of any later version published by the Free
Software Foundation.  If the Program does not specify a version number of
this License, you may choose any version ever published by the Free Software
Foundation.")
    (print "
  10. If you wish to incorporate parts of the Program into other free
programs whose distribution conditions are different, write to the author
to ask for permission.  For software which is copyrighted by the Free
Software Foundation, write to the Free Software Foundation; we sometimes
make exceptions for this.  Our decision will be guided by the two goals
of preserving the free status of all derivatives of our free software and
of promoting the sharing and reuse of software generally.
"))

  (defun warranty ()
    (print "			    NO WARRANTY

  11. BECAUSE THE PROGRAM IS LICENSED FREE OF CHARGE, THERE IS NO WARRANTY
FOR THE PROGRAM, TO THE EXTENT PERMITTED BY APPLICABLE LAW.  EXCEPT WHEN
OTHERWISE STATED IN WRITING THE COPYRIGHT HOLDERS AND/OR OTHER PARTIES
PROVIDE THE PROGRAM \"AS IS\" WITHOUT WARRANTY OF ANY KIND, EITHER EXPRESSED
OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.  THE ENTIRE RISK AS
TO THE QUALITY AND PERFORMANCE OF THE PROGRAM IS WITH YOU.  SHOULD THE
PROGRAM PROVE DEFECTIVE, YOU ASSUME THE COST OF ALL NECESSARY SERVICING,
REPAIR OR CORRECTION.

  12. IN NO EVENT UNLESS REQUIRED BY APPLICABLE LAW OR AGREED TO IN WRITING
WILL ANY COPYRIGHT HOLDER, OR ANY OTHER PARTY WHO MAY MODIFY AND/OR
REDISTRIBUTE THE PROGRAM AS PERMITTED ABOVE, BE LIABLE TO YOU FOR DAMAGES,
INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES ARISING
OUT OF THE USE OR INABILITY TO USE THE PROGRAM (INCLUDING BUT NOT LIMITED
TO LOSS OF DATA OR DATA BEING RENDERED INACCURATE OR LOSSES SUSTAINED BY
YOU OR THIRD PARTIES OR A FAILURE OF THE PROGRAM TO OPERATE WITH ANY OTHER
PROGRAMS), EVEN IF SUCH HOLDER OR OTHER PARTY HAS BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGES.
"))

  (export execute)
)